import argparse
from networks.network import Actor, Critic
from utils.utils import ReplayBuffer, make_mini_batch, convert_to_tensor

import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np

class PPO_MVP(nn.Module):
    def __init__(self, writer, device, state_dim, action_dim, args):
        super(PPO_MVP,self).__init__()
        self.args = args
        
        self.data = ReplayBuffer(action_prob_exist = True, max_size = self.args.traj_length, state_dim = state_dim, num_action = action_dim)
        self.actor = Actor(self.args.layer_num, state_dim, action_dim, self.args.hidden_dim, \
                           self.args.activation_function, self.args.last_activation, self.args.trainable_std)
        self.critic = Critic(self.args.layer_num, state_dim, 1, \
                             self.args.hidden_dim, self.args.activation_function, last_activation=None)
        
        self.actor_optimizer = optim.Adam(self.actor.parameters(), lr=self.args.actor_lr)
        self.critic_optimizer = optim.Adam(self.critic.parameters(), lr=self.args.critic_lr)

        self.writer = writer
        self.device = device
        self.episodes = None

        self.y = nn.Parameter(torch.zeros(1))
        self.y_optimizer = optim.Adam([self.y], lr=self.args.critic_lr)

    def get_action(self,x):
        mu,sigma = self.actor(x)
        return mu,sigma
    
    def v(self,x):
        return self.critic(x)
    
    def put_data(self,transition):
        self.data.put_data(transition)

    def put_episodes(self, episodes):
        self.episodes = episodes
        
    def get_gae(self, states, rewards, next_states, dones):
        values = self.v(states).detach()
        td_target = rewards + self.args.gamma * self.v(next_states) * (1 - dones)
        delta = td_target - values
        delta = delta.detach().cpu().numpy()
        advantage_lst = []
        advantage = 0.0
        for idx in reversed(range(len(delta))):
            if dones[idx] == 1:
                advantage = 0.0
            advantage = self.args.gamma * self.args.lambda_ * advantage + delta[idx][0]
            advantage_lst.append([advantage])
        advantage_lst.reverse()
        advantages = torch.tensor(advantage_lst, dtype=torch.float).to(self.device)
        return values, advantages
    
    def train_net(self, n_epi):
        data = self.data.sample(shuffle = False)
        states, actions, rewards, next_states, dones, old_log_probs = convert_to_tensor(self.device, data['state'], data['action'], data['reward'], data['next_state'], data['done'], data['log_prob'])
        
        old_values, advantages = self.get_gae(states, rewards, next_states, dones)
        returns = advantages + old_values
        advantages = (advantages - advantages.mean())/(advantages.std()+1e-3)
        
        assert self.args.train_epoch == 1

        for i in range(self.args.train_epoch):

            actor_loss_lst = []

            for state,action,old_log_prob,advantage,return_,old_value \
                in make_mini_batch(self.args.batch_size, states, actions, \
                                           old_log_probs,advantages,returns,old_values): 

                curr_mu,curr_sigma = self.get_action(state)
                value = self.v(state).float()
                curr_dist = torch.distributions.Normal(curr_mu,curr_sigma)
                entropy = curr_dist.entropy() * self.args.entropy_coef
                curr_log_prob = curr_dist.log_prob(action).sum(1,keepdim = True)

                #policy clipping
                ratio = torch.exp(curr_log_prob - old_log_prob.detach())
                surr1 = ratio * advantage
                surr2 = torch.clamp(ratio, 1-self.args.max_clip, 1+self.args.max_clip) * advantage
                actor_loss = (-torch.min(surr1, surr2) - entropy).mean()
                actor_loss_lst.append(actor_loss)
                
                #value clipping (PPO2 technic)
                old_value_clipped = old_value + (value - old_value).clamp(-self.args.max_clip,self.args.max_clip)
                value_loss = (value - return_.detach().float()).pow(2)
                value_loss_clipped = (old_value_clipped - return_.detach().float()).pow(2)
                critic_loss = 0.5 * self.args.critic_coef * torch.max(value_loss,value_loss_clipped).mean()
                
                # self.actor_optimizer.zero_grad()
                # actor_loss.backward()
                # nn.utils.clip_grad_norm_(self.actor.parameters(), self.args.max_grad_norm)
                # self.actor_optimizer.step()
                
                self.critic_optimizer.zero_grad()
                critic_loss.backward()
                nn.utils.clip_grad_norm_(self.critic.parameters(), self.args.max_grad_norm)
                self.critic_optimizer.step()
                
                if self.writer != None:
                    self.writer.add_scalar("loss/actor_loss", actor_loss.item(), n_epi)
                    self.writer.add_scalar("loss/critic_loss", critic_loss.item(), n_epi)


            ppo_loss = torch.stack(actor_loss_lst).mean()
            curr_y, M = self.prepare_mvp()
            policy_loss = 2 * curr_y * ppo_loss + M

            self.y_optimizer.step()

            self.actor_optimizer.zero_grad()
            policy_loss.backward()
            nn.utils.clip_grad_norm_(self.actor.parameters(), self.args.max_grad_norm)
            self.actor_optimizer.step()

    def prepare_mvp(self):
        episodes_num = len(self.episodes['state'])
        assert episodes_num == 1

        reward_lst = self.episodes['reward'][0]
        state_lst = self.episodes['state'][0]
        action_lst = self.episodes['action'][0]
        episode_length = len(reward_lst)

        # compute log_prob
        state = torch.tensor(np.array(state_lst)).float()
        action = torch.tensor(np.array(action_lst)).float()
        curr_mu, curr_sigma = self.get_action(state)
        curr_dist = torch.distributions.Normal(curr_mu,curr_sigma)
        curr_log_prob = curr_dist.log_prob(action).sum(1,keepdim = True)
        sum_log_prob = curr_log_prob.sum()

        # compute return
        G = torch.tensor(0).float()
        for R in reward_lst[::-1]:
            G = R + self.args.gamma * G


        curr_y = self.y.detach()
        M = G.pow(2) * sum_log_prob / episode_length

        y = self.y
        y_loss = (2* G + 1.0 / self.args.lam) * y - y.pow(2)
        y_loss = - y_loss
        self.y_optimizer.zero_grad()
        y_loss.backward()

        return curr_y, M

        
        



